/*
 **********************************************************************
 *     timer.c
 *     Copyright (C) 1999, 2000 Creative Labs, inc.
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include "hwaccess.h"

int sblive_timerIrqCallback(struct sblive_hw *sb_hw)
{
	struct emu_timer *timer;

	spin_lock(&sb_hw->timer_lock);

	timer = sb_hw->timer;
	while (timer != NULL)
	{
		queue_task(timer->task, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
		timer = timer->next;
	}

	spin_unlock(&sb_hw->timer_lock);

	return CTSTATUS_SUCCESS;
}

int sblive_timerinstall(struct sblive_hw *sb_hw, struct emu_timer **timer, struct tq_struct *task, unsigned delay)
{
	struct emu_timer *t;
	unsigned long flags;

	t = (struct emu_timer *) kmalloc(sizeof(struct emu_timer), GFP_KERNEL);
	if (t == NULL)
	{
		DPF(2, "timer alloc failed\n");
		return CTSTATUS_NOMEMORY;
	}

	DPD(3, "kmalloc: [%p]\n", t);

	if(delay < 1 )
		delay = 1;
	else if (delay >= 1024)
		delay = 1023;

	t->delay = delay;
	t->task = task;

	*timer = t;

	spin_lock_irqsave(&sb_hw->timer_lock, flags);

	if(sb_hw->timer == NULL)
	{
		sb_hw->timer = *timer;
		sb_hw->timer->next = NULL;
		sb_hw->timer_delay = delay;
		WRITE_FN0(sb_hw, TIMR_RATE, delay);
		sblive_irqmgrEnableIrq(sb_hw, INTE_INTERVALTIMERENB);
	}
	else
	{
		t = sb_hw->timer;
		sb_hw->timer = *timer;
		sb_hw->timer->next = t;
		if (delay < sb_hw->timer_delay)
		{
			sb_hw->timer_delay = delay;
			WRITE_FN0(sb_hw, TIMR_RATE, delay);
		}
	}

	spin_unlock_irqrestore(&sb_hw->timer_lock, flags);

	DPD(2, "timer rate --> %d\n", sb_hw->timer_delay);

	return CTSTATUS_SUCCESS;
}

int sblive_timeruninstall(struct sblive_hw *sb_hw, struct emu_timer *timer)
{
	struct emu_timer *t;
	unsigned delay;
	unsigned long flags;

	spin_lock_irqsave(&sb_hw->timer_lock, flags);

	t = sb_hw->timer;

	if( t == NULL){
		spin_unlock_irqrestore(&sb_hw->timer_lock, flags);
		return CTSTATUS_SUCCESS;
	}

	delay = timer->delay;

	if(sb_hw->timer == timer)
	{
		sb_hw->timer = timer->next;
	}
	else
	{
		while(t->next != timer)
			t = t->next;

		t->next = timer->next;
	}

	kfree(timer);

	DPD(3, "kfree: [%p]\n", timer);

	if(sb_hw->timer == NULL)
	{
		sblive_irqmgrDisableIrq(sb_hw, INTE_INTERVALTIMERENB);
	}
	else if (delay == sb_hw->timer_delay)
	{
		t = sb_hw->timer;
		sb_hw->timer_delay = 1024;

		while(t != NULL){

			if(t->delay < sb_hw->timer_delay)
				sb_hw->timer_delay = t->delay;
			t = t->next;
		}

		WRITE_FN0(sb_hw, TIMR_RATE, sb_hw->timer_delay);

		DPD(2, "timer rate --> %d\n", sb_hw->timer_delay);
	}

	spin_unlock_irqrestore(&sb_hw->timer_lock, flags);

	return CTSTATUS_SUCCESS;
}

int sblive_timerinit(struct sblive_hw *sb_hw)
{
	sb_hw->timer = NULL;
	sb_hw->timer_lock = SPIN_LOCK_UNLOCKED;

	return CTSTATUS_SUCCESS;
}
